#property copyright "Copyright 2022, Anil Varma"
#property link      "AnilHVarma@gmail.com"
#property version   "1.00"

#include "../MQL5 Forum/IndicatorBaseMQL5.mqh"

//+-----------------------------------------------------------------------------------------------------------------------------+
//| CLASS:				CFractalsMQL5.mqh
//| APPLICATION:	CLASS FOR STOCHASTIC OSCILLATOR
//+-----------------------------------------------------------------------------------------------------------------------------+
class CFractalsMQL5 final : public CIndicatorBaseMQL5 {

public:

  CFractalsMQL5() : CIndicatorBaseMQL5() { }
  CFractalsMQL5(string pSymbol,ENUM_TIMEFRAMES pTimeFrame) : CIndicatorBaseMQL5() { InitHandle(pSymbol,pTimeFrame); }
 ~CFractalsMQL5();

private:

	int					mLookBack;																	// MAXIMUM NO OF BARS TO LOOK BACK FOR FRACTALS
  int         mHandle;

  // METHOD TO GET VALUE OF EACH BAR IN INDICATOR OnCalculate()
	double 			GetIndexFractalPrice(int pIndicatorBuffer,int pIndex);	// INTERNAL CALCULATION

public:
	void				InitHandle(string pSymbol,ENUM_TIMEFRAMES pTimeFrame);
	//+---------------------------------------------------------------------------------------------------------------------------+
	//| METHODS TO RETURN INDICATOR VALUES TO CALLING EA / CLASS
	//+---------------------------------------------------------------------------------------------------------------------------+
	void 				GetArrayBEARISH(int pFractalCount,SFractal &pFractalBearish[]);
	void 				GetArrayBULLISH(int pFractalCount,SFractal &pFractalBullish[]);
	//+---------------------------------------------------------------------------------------------------------------------------+
	//| METHODS TO CALCULATE IF FRACTALS RISING OR FALLING
	//+---------------------------------------------------------------------------------------------------------------------------+
private:
	SFractal		mFractalBull[];
	SFractal		mFractalBear[];
public:
	bool				IsFractalTrend_LONG();
	bool				IsFractalTrend_Short();
	//+---------------------------------------------------------------------------------------------------------------------------+
	//| METHODS TO IDENTIFY IF FRACTAL TREND WEAKENING
	//+---------------------------------------------------------------------------------------------------------------------------+
		bool			IsLongTrendExhaust();
		bool			IsShortTrendExhaust();

}; // END Of Class Definition
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:       ~CFractalsMQL5()
//| APPLICATION:  DECONSTRUCTOR METHOD
//+-----------------------------------------------------------------------------------------------------------------------------+
CFractalsMQL5::~CFractalsMQL5() {

  IndicatorRelease(mHandle);

} // END Of default deconstructor ~CFractalsMQL5()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:				InitHandle()
//| APPLICATION:	METHOD TO INITIALIAZE MEMBER VARIABLES OF THE INDICATOR
//+-----------------------------------------------------------------------------------------------------------------------------+
void CFractalsMQL5::InitHandle(string pSymbol,ENUM_TIMEFRAMES pTimeFrame) {

	// ASSIGN SYMBOL AND TIME FRAME
	mSymbol		 = pSymbol;
	mTimeFrame = pTimeFrame;

	// INITIALIAZE VARIABLES
	mLookBack	 = 2000;																		// LOOKBACK MAXIMUM 'n' BARS TO FIND REQUIRED NO OF FRACTALS

	// DECLARE INDICATOR HANDLES
	TesterHideIndicators(true);

	mHandle = iFractals(mSymbol,mTimeFrame);
  if(mHandle == INVALID_HANDLE) {
		ResetLastError();
    PrintFormat("[%s] %s: Error[#%i] creating handle iFractals",EnumToString(mTimeFrame),(string)__FUNCTION__,GetLastError());
  }
  
	TesterHideIndicators(false);

} // END Of method InitHandle()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:  			GetArrayBEARISH()
//| APPLICATION:  METHOD TO RETURN UPPER iFractals INDEX / VALUES INTO PASSED IN STRUCTIRE ARRAY
//+-----------------------------------------------------------------------------------------------------------------------------+
void CFractalsMQL5::GetArrayBEARISH(int pFractalCount,SFractal &pFractalBearish[]) {

	// IF HANDLES NOT VALID, REINITIALIZE THEM
  if(mHandle == INVALID_HANDLE) {
    InitHandle(mSymbol,mTimeFrame);
  }

	int countMax = pFractalCount;															// SET MAXIMUM NUMBER OF FRACTALS TO LOOK FOR UPPER LINE
	ZeroMemory(pFractalBearish);
	ArraySetAsSeries(pFractalBearish,true);
	ArrayResize(pFractalBearish,countMax+1);

	int count = 1;																						// FIRST FRACTAL INDEX WILL BE ONE, THEN INCREMENTED WITH COUNTER++

	// LOOP FOR MAXIMUM mLookBars TO FIND IF FRACTAL EXIST
  for(int i = 0; i <= mLookBack; i++) {

		double vPrice = GetIndexFractalPrice(UPPER_LINE,i);
		if(vPrice != EMPTY_VALUE && count <= countMax) {
			pFractalBearish[count].idx 	 = i;
			pFractalBearish[count].price = vPrice;
			count++;
		}

		if(count > countMax)																		// EXIT, WHEN REQUIRED NO OF FRACTALS FOUND
			return;
  }

} // END Of method GetArrayBEARISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD:  			GetArrayBULLISH()
//| APPLICATION:  METHOD TO RETURN LOWER iFractals INDEX / VALUES INTO PASSED IN STRUCTIRE ARRAY
//+-----------------------------------------------------------------------------------------------------------------------------+
void CFractalsMQL5::GetArrayBULLISH(int pFractalCount,SFractal &pFractalBullish[]) {

	// IF HANDLES NOT VALID, REINITIALIZE THEM
  if(mHandle == INVALID_HANDLE) {
    InitHandle(mSymbol,mTimeFrame);
  }

	int countMax = pFractalCount;															// SET MAXIMUM NUMBER OF FRACTALS TO LOOK FOR UPPER LINE
	ZeroMemory(pFractalBullish);
	ArraySetAsSeries(pFractalBullish,true);
	ArrayResize(pFractalBullish,countMax+1);

	int count = 1;																						// FIRST FRACTAL INDEX WILL BE ONE, THEN INCREMENTED WITH COUNTER++

	// LOOP FOR MAXIMUM mLookBars TO FIND IF FRACTAL EXIST
  for(int i = 0; i < mLookBack; i++) {

		double vPrice = GetIndexFractalPrice(LOWER_LINE,i);
		if(vPrice != EMPTY_VALUE && count <= countMax) {
			pFractalBullish[count].idx 	 = i;
			pFractalBullish[count].price = vPrice;
			count++;
		}

		if(count > countMax)																		// EXIT, WHEN REQUIRED NO OF FRACTALS FOUND
			return;
  }

} // END Of method GetArrayBULLISH()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  GetIndexFractalPrice()
//| APPLICATION:  RETURNS FRACTAL PRICE VALUE ON SPECIFIED FRACTAL BUFFER AND INDEX
//+-----------------------------------------------------------------------------------------------------------------------------+
double CFractalsMQL5::GetIndexFractalPrice(int pIndicatorBuffer,int pIndex) {

  double bufferData[1];
  if(CopyBuffer(mHandle,pIndicatorBuffer,pIndex,1,bufferData) < 1) {
		ResetLastError();
    PrintFormat("[%s] %s: Error[#%i] getting iFractals BUFFER[%i] at Index[%i]",EnumToString(mTimeFrame),string(__FUNCTION__),
    						pIndicatorBuffer,GetLastError(),pIndex);
    return(0.0);
  }

  return(bufferData[0]);

} // END Of method GetIndex()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsFractalTrend_LONG()
//| APPLICATION:  RETURNS IF BULLISH FRACTAL PRICES FORMING LOWER HIGHs OR SAY UP TREND)
//| NOTE:					USE FOR LONG ENTRY OR SHORT EXIT
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CFractalsMQL5::IsFractalTrend_LONG(void) {

	int IdxGap = 5;																					// TWO BULL FRACTALS SHOULD BE AWAY AT LEAST FIVE BARS FROM EACH OTHER
	GetArrayBULLISH(3,mFractalBull);
	GetArrayBEARISH(3,mFractalBear);

	if(!(mFractalBear[3].price < mFractalBear[2].price && mFractalBear[2].price < mFractalBear[1].price)) 					return(false);
	if(!(mFractalBull[3].price < mFractalBull[2].price && mFractalBull[2].price < mFractalBull[1].price)) 					return(false);
	if(!((mFractalBull[3].idx - mFractalBull[2].idx) >= IdxGap))																										return(false);
	if((mFractalBull[2].idx - mFractalBull[1].idx) >= IdxGap)																												return(true);

	return(false);
/*
	if(mFractalBull[3].price < mFractalBull[2].price && mFractalBull[2].price < mFractalBull[1].price) 					return(true);
	else if(mFractalBull[3].price < mFractalBull[1].price) 																											return(true);
	else if(mFractalBull[2].price < mFractalBull[1].price) 																											return(true);
	else 																																																				return(false);
*/
} // END Of method IsFractalTrend_LONG()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsFractalTrend_Short()
//| APPLICATION:  RETURNS IF BEARISH FRACTAL PRICES FORMING HIGHER LOWs OR TO SAY DOWN TREND
//| NOTE:					USE FOR SHORT ENTRY OR LONG EXIT
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CFractalsMQL5::IsFractalTrend_Short(void) {

	int IdxGap = 5;																					// TWO BEAR FRACTALS SHOULD BE AWAY AT LEAST FIVE BARS FROM EACH OTHER
	GetArrayBEARISH(3,mFractalBear);
	GetArrayBULLISH(3,mFractalBull);

	if(!(mFractalBear[3].price > mFractalBear[2].price && mFractalBear[2].price > mFractalBear[1].price)) 					return(false);
	if(!(mFractalBull[3].price > mFractalBull[2].price && mFractalBull[2].price > mFractalBull[1].price)) 					return(false);
	if(!((mFractalBear[3].idx - mFractalBear[2].idx) >= IdxGap))																										return(false);
	if((mFractalBear[2].idx - mFractalBear[1].idx) >= IdxGap)																												return(true);

	return(false);
/*	
	if(mFractalBear[3].price > mFractalBear[2].price && mFractalBear[2].price > mFractalBear[1].price) 							return(true);
	else if(mFractalBear[3].price > mFractalBear[1].price) 																													return(true);
	else if(mFractalBear[2].price > mFractalBear[1].price) 																													return(true);
	else 																																																						return(false);
*/

} // END Of method IsFractalTrend_Short()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsLongTrendExhaust()
//| APPLICATION:  RETURNS [true] IF DISTANCE BETWEEN RISING BEARISH FRACTAL IS DECREASING, A SIGN OF TREND WEAKENING
//| NOTE:					TREND WILL HAVE LONGER DISTANCE BETWEEN FRACTALS THEN TREND WEAKENING. WE DEFAULT DISTANCE [2][1] TO BE MORE
//|								THEN TWO TIMES OF DISTANCE AND SHOULD BE AT LEAST 8 BARS BETWEEN [3][2]
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CFractalsMQL5::IsLongTrendExhaust(void) {

	GetArrayBEARISH(3,mFractalBear);

	//PrintFormat("%s: [%s] FBear[3][%.1f] < FBear[2][%.1f] < FBear[1][%.1f]",(string)__FUNCTION__,EnumToString(mTimeFrame),mFractalBear[3].price,mFractalBear[2].price,mFractalBear[1].price);
	if(!(mFractalBear[3].price < mFractalBear[2].price && mFractalBear[2].price < mFractalBear[1].price))						return(false);
	int diffUpper = mFractalBear[3].idx - mFractalBear[2].idx;

	//PrintFormat("%s: [%s] diffIdx Upper[%.2f]",(string)__FUNCTION__,EnumToString(mTimeFrame),diffUpper);
	if(!(diffUpper >= 8))																																														return(false);

	// NOTE: IN MQL int / int RETURNS int VARIABLE. WE NEED TO SPECIFICALLY CAST IT IF NEED RETURN SAY IN double VARIABLE
	int diffLower = mFractalBear[2].idx - mFractalBear[1].idx;
	//PrintFormat("%s: [%s] diffIdx Upper[%.2f] / Lower[%.2f] = [%.2f] > 2.00",(string)__FUNCTION__,EnumToString(mTimeFrame),diffUpper,diffLower,((double)diffUpper / diffLower));
	if(((double)diffUpper / diffLower) > 2.00)																																			return(true);
	else 																																																						return(false);

} // END Of method IsLongTrendExhaust()
//+-----------------------------------------------------------------------------------------------------------------------------+
//| METHOD BODY:  IsShortTrendExhaust()
//| APPLICATION:  RETURNS [true] IF DISTANCE BETWEEN FALLING BULLISH FRACTAL IS DECREASING, A SIGN OF TREND WEAKENING
//+-----------------------------------------------------------------------------------------------------------------------------+
bool CFractalsMQL5::IsShortTrendExhaust(void) {

	GetArrayBULLISH(3,mFractalBull);

	if(!(mFractalBull[3].price > mFractalBull[2].price && mFractalBull[2].price > mFractalBull[1].price))						return(false);
	int diffUpper = mFractalBull[3].idx - mFractalBull[2].idx;
	if(!(diffUpper >= 8))																																														return(false);

	// NOTE: IN MQL int / int RETURNS int VARIABLE. WE NEED TO SPECIFICALLY CAST IT IF NEED RETURN SAY IN double VARIABLE
	int diffLower = mFractalBull[2].idx - mFractalBull[1].idx;
	if(((double)diffUpper / diffLower) > 2.00)																																			return(true);
	else 																																																						return(false);

} // END Of method IsShortTrendExhaust()
//+-----------------------------------------------------------------------------------------------------------------------------+
